/* This is a driver for a usbsm graphics tablet.
 * It was based on my PCIbase driver, from
 *  http://www.bolthole.com/solaris/drivers/
 *
 *
 *  usbsm.c 1.17 
 */

#define _SYS_VARARGS_H         /* sys/varargs.h hack for gcc2 */
#include <stdarg.h>

#include <sys/usb/usba.h>
  /* we dont actually use the hid routines.. we use the defines */
#include <sys/usb/clients/hid/hid.h>
#include <sys/msio.h>

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/file.h>

#include <sys/vuid_event.h>

#include <sys/stream.h>
#include <stropts.h>
#include <sys/signal.h>

/* THESE MUST BE LAST */
#include <sys/ddi.h>
#include <sys/sunddi.h>


/* As well as changing the NAME of this, you will also want to change
 * the TYPE of the minor node created.
 */
#define DEVNODE_NAME	"usbsm"	
#define DEVNODE_TYPE	DDI_NT_SERIAL /* see sunddi.h */


/**********************************************************************/

/* This ASSUMES you create minor nodes where the higher bits denote
 * the instance of the driver. (aka "which card is this?")
 * The lower bits denote clone number, or anything else you want.
 */

#define DEVT_TO_INST(devt) (getminor(devt)>>4)
#define DEVT_TO_CLONENUM(devt) (getminor(devt) & 0xf)

/* Used to choose a node  num for a device node compatible with DEVT_TO_INST*/
#define INST_TO_NODENUM(instance) (instance<<4)


/*Our globally visible stuffs, that we populate the global
 * driver entry structs with  [base_cb_ops, base_ops]
 * Yeah, its kind of wierd to have a "static" be also externally visible,
 * but that's traditionally how these are declared.
 */
static	int	usbsm_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);

static	int	usbsm_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);

static	int	usbsm_open(queue_t *q,dev_t *dev, int ofl, int otyp, cred_t *c);
static	int	usbsm_wput(queue_t *q,mblk_t *mblk);

static	int	usbsm_rput(queue_t *q,mblk_t *mp);
static	int	usbsm_rsrv(queue_t *q);

static	int	usbsm_close(queue_t *q, int flag, cred_t *cred_p);

static	int	usbsm_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);


static int debug_level;


/**********************************************************************/
/* STREAMS driver info stuff  */

static struct module_info stream_minfo = {
	0x0001,		/* module id number (?) */
	"usbsm",	/* module name */
	0,		/* min packet size */
	INFPSZ,		/* max packet size */
	512,		/* "high water mark" (?)*/
	128,		/* "low water mark" (?)*/
};

/* Queue for reading data */
static struct qinit stream_rinit = {
	usbsm_rput, /* qi_putp */
	usbsm_rsrv, /* qi_srvp */
	usbsm_open,
	usbsm_close,
	NULL, /* qi_qadmin */
	&stream_minfo,
	NULL /* qi_mstat */
};

/* Queue for writing data (to the dev) */
static struct qinit stream_winit = {
	/* Im not sure we should really have ANYTHING here.. */
	usbsm_wput, /* qi_put */ 
	NULL, /* qr_srv not needed */
	NULL, /* qi_qopen */
	NULL, /* qi_qclose */
	NULL, /* qi_qadmin */
	&stream_minfo,
	NULL /* qi_mstat */
};

static struct streamtab stream_rwinfo = {
	&stream_rinit,
	&stream_winit,
	NULL,		/* st_muxrinit */
	NULL		/* st_muxwinit */
};

/* end of STREAMS specific struct stuff  */
/**********************************************************************/



/* "char/block operations" */
/* Since we are a STREAMS dev, all the stuff that is NORMALLY referenced in 
 * here, is now referenced from the wierd STREAMs specific subtree.
 */
static struct cb_ops	usbsm_cb_ops = {
	nulldev,		/* open() */
	nulldev,		/* close() */
	nodev,			/* strategy()		*/
	nodev,			/* print routine	*/
	nodev,			/* no dump routine	*/
	nodev,			/* read() */
	nodev,			/* write() */
	nodev,			/* ioctl */
	nodev,			/* no devmap routine	*/
	nodev,			/* no mmap routine	*/
	nodev,			/* no segmap routine	*/
	nochpoll,		/* no chpoll routine	*/
	ddi_prop_op,
	&stream_rwinfo,		/* cb_stream struct */
	D_NEW | D_MP | D_MTPERMOD , /* safe for multi-thread/multi-processor */
	CB_REV,			/* cb_ops version? */
	nodev,			/* cb_aread() */
	nodev,			/* cb_awrite() */
};

/* "device operations". */
static struct dev_ops usbsm_ops = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	usbsm_getinfo,		/* devo_getinfo */
	nulldev,		/* devo_identify */
	nulldev,		/* devo_probe */
	usbsm_attach,		/* devo_attach */
	usbsm_detach,		/* devo_detach */
	nodev,			/* devo_reset */
	&usbsm_cb_ops,		/* devo_cb_ops */
	(struct bus_ops *)0,	/* devo_bus_ops */
	NULL,			/* devo_power */
};

/* Ignore this.
 * It is already declared in <sys/modctl.h>
 * It is only a filler for the next struct.
 */
extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"BOLTusbsm wheelmouse v2.1",
	&usbsm_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 is indicated by manual */
	{&modldrv,NULL,NULL,NULL}

};


/************************************************************
 * And here's finally the vars specific to THIS driver,
 * that get allocated on a per-instance basis.
 * Here's where you should keep instance-specific mutexes,
 * DMA handles, register mapping handles, and stuff like that
 *
 ************************************************************/


#define MAX_INSTANCES 4 /* max of 4 hardware devices loaded */

/* 
 * Note that this mutex gets used in _init. Which means this mutex
 * isn't just for this "instance" of the driver. It is for ALL INSTANCES.
 * If you have multiple cards, you have multiple instances of the driver,
 * with this single shared mutex across them all
 */
static kmutex_t   usbsm_global_mutex;

/* This is SORT OF like	int (*array)[] , except you NEVER USE THIS DIRECTLY.
 * You just use it with ddi_get_soft_state().
 * 
 *    "Global Softstate Handle" 
 */
void * usbsm_glob_soft_handle;

/* This gets filled out on a per-instance basis by usbsm_attach */
struct usbsm_softstate {
	dev_info_t *dip; /* you always need this cached */
	kmutex_t instance_mutex;

	char lastdata[8];

	usb_config_descr_t usb_config;
	usb_endpoint_descr_t endpoint;
	int usb_interfacenum;
	usb_interface_descr_t usb_interface_d;
	int openflags; /* tell if we have been open()ed or not */
	usb_pipe_handle_t usb_pipe_handle;

	size_t max_transfer_size; /* technically, BULK max transfer... */
	int packetsize;   /* expected size of usb data packet from mouse */
	int packetoffset; /* Some mice need "tweaking" */

	queue_t *rq, *wq;

	/* toggle, to see if we need to set VUID format data, or "raw" */
	int vuid_format; 
};

/* defines used on the "openflags" var in softstate*/
#define USBMS_OPEN_HELD		0x01
#define USBMS_OPEN_ENDING	0x02
#define USBMS_OPEN_SHUTDOWN	0x03 /* flag "shutting down" case */


/*****************************************************************/
/*****************************************************************/
/*              Start of actual kernel routines here             */
/*****************************************************************/
/*****************************************************************/



void
usbsm_debugprint(const char *format, ...)
{
	va_list args;
	va_start(args, format);

	if(debug_level==0){
		return;
	}

	/* DO NOT USE cmn_err here!!! it will not be interrupt-safe */
	vprintf(format, args);
	va_end(args);
}



/*Like the usbsm_global_mutex above, this routine is GLOBAL across all
 * instances.
 * It actually gets called when you add_drv your driver from the command
 * line, before  * it is hooked up to any specific hardware.
 * See usbsm_attach() below for instance-specific stuff.
 *
 * HOWEVER... Its often nice to pre-allocate "softstate" space
 * here, if you dont mind hardcoding the total number of possible
 * instances you can have. So we do that.
 * 
 */
int
_init(void)
{
       int     ret;

	/*printf("usbsm: _init called\n");*/
	/*
               * Initialize global mutex before mod_install'ing driver.
               * If mod_install() fails, must clean up mutex initialization
               */
	mutex_init(&usbsm_global_mutex, "usbsm: Global Mutex",
                     MUTEX_DRIVER, (void *)NULL);

	/* This is kind of like
	 * malloc(sizeof(struct usbsm_softstate *MAX_INSTANCES))
	 */
	ret=ddi_soft_state_init(&usbsm_glob_soft_handle,
				sizeof(struct usbsm_softstate),
			        MAX_INSTANCES);
	if(ret!=0){
		return (ret);
	}
	
	if ((ret = mod_install(&modlinkage)) != 0) {
		mutex_destroy(&usbsm_global_mutex);
		ddi_soft_state_fini(&usbsm_glob_soft_handle);
		cmn_err(CE_WARN, "usbsm: _init: mod_install FAILED:\n");
		return DDI_FAILURE;
	}

	usbsm_debugprint("usbsm: _init returning SUCCESS\n");
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	
	usbsm_debugprint("usbsm: _info called\n");
	return (mod_info(&modlinkage, modinfop));
}

/* This is the match to the _init() routine. Clean up stuff allocated
 * specifically in _init() here.
 * Stuff in attach() should be cleaned up in detach()
 */
int
_fini(void)
{
	int ret;

	usbsm_debugprint("usbsm: _fini called\n");
	
        /*
         * If mod_remove() is successful, we destroy our global mutex
         */
	if ((ret = mod_remove(&modlinkage)) == 0) {
		ddi_soft_state_fini(&usbsm_glob_soft_handle);
		mutex_destroy(&usbsm_global_mutex);
	}
	return (ret);
}

/************************************************************/

/* Take "Firm_event",and  put it on the internal STREAMS message queue,
 * to eventually be read by the user process
 */
static void send_firm_event(queue_t *q, Firm_event *fptr)
{
	mblk_t *newmptr;
	newmptr=allocb(sizeof(Firm_event), 0);
	bcopy(fptr, newmptr->b_wptr, sizeof(Firm_event));
	newmptr->b_wptr += sizeof(Firm_event);
/****
	usbsm_debugprint("have rput event:rptr=%p, wptr=%p\n",
	       newmptr->b_rptr,newmptr->b_wptr);
	usbsm_debugprint(" extended info: r=%p, w=%p",
	       newmptr->b_datap->db_mblk->b_rptr,
	       newmptr->b_datap->db_mblk->b_wptr);
****/
	usbsm_rput(q, newmptr);
}



/* need to convert the "raw" USB mouse event, into sun style
 * "VUID events".
 * Raw USB mouse events are quads, defaulted to relative data:
 * byte 0=button, 1=x, 2=y, 3=scroll
 *
 * Note that sun vuid wants   (up=positivey, right=positivex)
 * whereas USB seems to have  (DOWN=positivey, right=positivex)
 *
 * This may generate more than one VUID event for a single 'raw' event.
 *
 */
static void
vuid_encode_event(struct usbsm_softstate *softstate,mblk_t*mblkptr)
{
	char *data=(char*)mblkptr->b_rptr;
	int datalen=mblkptr->b_wptr - mblkptr->b_rptr;
	Firm_event fevent;

	if(datalen!=softstate->packetsize){
		printf("DEBUG: vuid_encode_format ignoring data of len %d\n",
		       datalen);
		return;
	}

	if(softstate->packetoffset!=0){
		data=&data[softstate->packetoffset];
	}

	bzero(&fevent, sizeof(fevent));

	uniqtime32(&fevent.time);


	usbsm_debugprint("vuid_encode: raw data= %d:%d:%d:%d\n",
	                 data[0],data[1],data[2],data[3]);

	/* I dont know why I set FE_PAIR_ABSOLUTE, below.
	 * I just know that according to testvuid, the sun driver does it
	 */

	/* check for location (X/Y) changes */
	if(data[1] != 0){
		/* X change */

		fevent.id=LOC_X_DELTA;
		fevent.value=data[1];
		fevent.pair_type=FE_PAIR_ABSOLUTE;
		fevent.pair=LOC_X_ABSOLUTE;

		send_firm_event(softstate->rq, &fevent);
		fevent.time.tv_usec++;
	}
	if(data[2] != 0){
		/* X change */

		fevent.id=LOC_Y_DELTA;
		/* Have to flip value. See notes at top of function */
		fevent.value=0 - data[2];
		fevent.pair_type=FE_PAIR_ABSOLUTE;
		fevent.pair=LOC_Y_ABSOLUTE;

		send_firm_event(softstate->rq, &fevent);
		fevent.time.tv_usec++;
	}
	
	fevent.pair_type=FE_PAIR_NONE;


#define LEFTBUTMASK 0x01
#define MIDDLEBUTMASK 0x04
#define RIGHTBUTMASK 0x02



	/* check for button changes */
	/* Assumes that only one button can change state with a
	 * single USB event!
	 */
	if(data[0] != softstate->lastdata[0]){
		if((data[0]& LEFTBUTMASK) != (softstate->lastdata[0] &LEFTBUTMASK)){
			fevent.id=BUT(1);
			if(data[0] & LEFTBUTMASK){
				fevent.value=1;
			} else {
				fevent.value=0;
			}
			send_firm_event(softstate->rq, &fevent);
			fevent.time.tv_usec++;

		}
		if((data[0]& MIDDLEBUTMASK) != (softstate->lastdata[0] &MIDDLEBUTMASK)){
			fevent.id=BUT(2);
			if(data[0] & MIDDLEBUTMASK){
				fevent.value=1;
			} else {
				fevent.value=0;
			}
			send_firm_event(softstate->rq, &fevent);
			fevent.time.tv_usec++;

		}
		if((data[0]& RIGHTBUTMASK) != (softstate->lastdata[0] &RIGHTBUTMASK)){
			fevent.id=BUT(3);
			if(data[0] & RIGHTBUTMASK){
				fevent.value=1;
			} else {
				fevent.value=0;
			}
			send_firm_event(softstate->rq, &fevent);
			fevent.time.tv_usec++;

		}
	}

	/* The packetsize check, is to try to allow use by non-wheeled mice.
	 * Untested whether it will work.
	 * Otherwise, this block just uses the scrollwheel data.
	 */
	if((softstate->packetsize>=4) && (data[3] != 0)){
		int clickcount=0;
		/* convert scroll motion, into button4/button5 "events */
		/* "up" is >0, "down" is <0 */
		if(data[3] > 0){
			fevent.id=BUT(4);
			clickcount=data[3]&0x3f;
		} else {
			fevent.id=BUT(5);
			clickcount=(0-data[3]) &0x3f;
		}
		while(clickcount-->0){
			fevent.value=1;
			send_firm_event(softstate->rq, &fevent);
			fevent.time.tv_usec++;

			/* have to fake button"release" event also */
			fevent.value=0;
			send_firm_event(softstate->rq, &fevent);
			fevent.time.tv_usec++;
		}

	}

	bcopy(data, softstate->lastdata, 4);

}


/**********************************************************************/
/* This is probably interrupt context. 
 * mblk_t is a STREAMS style Message Block. Arrg.
 */
static int
usbsm_pipe_callback(usb_pipe_handle_t pipe, usb_opaque_t arg,mblk_t*mblkptr)
{
	struct usbsm_softstate *softstate=(struct usbsm_softstate*)arg;

	if(debug_level>1){
		usbsm_debugprint("usbsm_pipe_callback called\n");
	}

	if(mblkptr==NULL){
		printf("mblk==NULL\n");
		return (USB_SUCCESS);
	}

	if(softstate->rq==NULL){
		printf("ERROR in callback: softstate->rq==NULL?\n");
		return USB_FAILURE;
	}

	if(debug_level>=2){
		int datalen;
		datalen=mblkptr->b_wptr - mblkptr->b_rptr;
		usbsm_debugprint(
		  "^usbsm_pipe_callback - calling usbsm_rput with %d bytes\n",
		  datalen);
	}

	if(softstate->vuid_format){
		vuid_encode_event(softstate, mblkptr);
		freemsg(mblkptr);
	} else {
		/* just shove raw data back to user */
		usbsm_rput(softstate->rq, mblkptr);
	}

	return (USB_SUCCESS);
}

/* temp test routine, that is only used for test purposes, essentially.
 */
static void
usbsm_test_callback(usb_opaque_t arg,int arg1, uint_t arg2)
{
	usbsm_debugprint("usbsm_TEST_callback called\n");
	return;
}

/* This gets called for partial USB transfers.
 * EG: some error occurred.
 */
static int
usbsm_exception_callback(usb_pipe_handle_t pipe, usb_opaque_t arg,
			 uint_t complete_reason,mblk_t*mblkptr,uint_t flag)
{
	/*struct usbsm_softstate *softstate=(struct usbsm_softstate*)arg;*/

	usbsm_debugprint("usbsm_EXCEPTION_callback called: reason %d\n",
			 complete_reason);
	return (USB_SUCCESS);
}

static int usbsm_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
			     void *arg, void **result){
	struct usbsm_softstate *softstate;
	int returnval=DDI_FAILURE;
	dev_t devt=(dev_t)arg;
	int instance=DEVT_TO_INST(devt);
	
	usbsm_debugprint("usbsm: getinfo called\n");
	
	switch(infocmd){
	    case DDI_INFO_DEVT2DEVINFO:
		softstate=ddi_get_soft_state(usbsm_glob_soft_handle,instance);
		if(softstate==NULL){
			*result=NULL;
			break;
		}
		*result=softstate->dip;
		returnval=DDI_SUCCESS;
		break;
	    case DDI_INFO_DEVT2INSTANCE:
		*result=(void*)instance;
		returnval=DDI_SUCCESS;
		break;
	}

	return (returnval);
}




/* This routine can be called by usbsm_attach (if errors occur)
 * and usbsm_detach (on normal detach)
 * Idea is to clean up any mutexes, register mappings, etc. nicely.
 * We also free the instance-softstate.
 * We do NOT remove the minor node, however. Only detach does
 * that currently.
 */
static void usbsm_cleanup(struct usbsm_softstate *softstate){
	int instance;

	if(softstate==NULL){
		/* yes, paranoia is good*/
		cmn_err(CE_WARN,"ERROR: usbsm_cleanup called with NULL");
		return;
	}
   	instance=ddi_get_instance(softstate->dip);

	mutex_destroy(&softstate->instance_mutex);

	ddi_soft_state_free(usbsm_glob_soft_handle,instance);
}


#define REQUEST_TYPE USB_DEV_REQ_RECIPIENT_INTERFACE|USB_DEV_REQ_TYPE_CLASS
#define REQUEST_INDEX softstate->usb_interface_d.bInterfaceNumber
#define REQUEST_CMD USB_REQ_SET_CONFIGURATION

/*
 * _init() has already been called. This routine gets called
 * once for EVERY MATCHING PIECE OF HARDWARE.
 * It gets called for initial "attach", and also for resume(/suspend)
 * You would usually allocate space for state variables, and
 * map device registers in here. We dont have registers to map, however.
 *
 * To avoid reboot loops, be very very careful putting things in here.
 * If you want to put extra debug type stuff in, it's safer putting
 * it in usbsm_open() or someplace like that, that gets called after
 * the driver is safely loaded
 *
 * Be sure to free/deallocate anything you create here, in
 *  usbsm_detach()
 */
static int usbsm_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
   	int instance;
	struct usbsm_softstate *softstate;
	size_t usb_raw_len;
	uchar_t *usb_rawconf;

	/* default debug level to **0**
	 * You can override in the .conf file.
	 */
	debug_level=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
                                    "debug-level", 0);

	usbsm_debugprint(" usbsm_attach called\n");

   	switch(cmd){
	 case DDI_RESUME:
	   	cmn_err(CE_NOTE, "usbsm: Ignoring attach_RESUME\n");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		cmn_err(CE_NOTE, "usbsm: Ignoring attach_PM_RESUME\n");
		return DDI_FAILURE;
	 case DDI_ATTACH:
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "usbsm: unknown attach command\n");
	   	return DDI_FAILURE;
	}
   	instance=ddi_get_instance(dip);
   	usbsm_debugprint("usbsm: attach for instance %d\n", instance);


	/* We dont have to check for being over MAX_INSTANCES, because
	 * these two calls do it implicitly
	 */
	if (ddi_soft_state_zalloc(usbsm_glob_soft_handle, instance) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "usbsm: %d: Failed to alloc soft state",
		    instance);
		return (DDI_FAILURE);
	}
	softstate=ddi_get_soft_state(usbsm_glob_soft_handle,instance);
	if(softstate==NULL){
		cmn_err(CE_WARN, "usbsm: attach for instance %d: Failed to get soft state",
			instance);
		ddi_soft_state_free(usbsm_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}


	softstate->packetsize=ddi_prop_get_int(DDI_DEV_T_ANY,dip,
	                                       DDI_PROP_DONTPASS,
                                               "packetsize", 4);

	softstate->packetoffset=ddi_prop_get_int(DDI_DEV_T_ANY,dip,
	                                       DDI_PROP_DONTPASS,
                                               "packetoffset", 0);

	/* Note: 3 is a *special case*. We normally insist on a
	 * packetsize of 4, for use of the scrollwheel
	 */
	if(softstate->packetsize <3){
		cmn_err(CE_WARN, "usbsm: invalid packetsize %d\n",
			softstate->packetsize);
		ddi_soft_state_free(usbsm_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}
	if((softstate->packetsize - softstate->packetoffset)<4){
		cmn_err(CE_WARN, "usbsm: packetoffset too small\n");
		ddi_soft_state_free(usbsm_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}

	softstate->dip = dip;
	softstate->openflags = 0; /* redundant, but what the heck */
	

	mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER, NULL);


	usb_rawconf = usb_get_raw_config_data(dip,&usb_raw_len);

	/* Fill out "configuration" descriptor */
	if(usb_parse_configuration_descr(usb_rawconf, usb_raw_len,
					&softstate->usb_config,
					USB_CONF_DESCR_SIZE) !=
						USB_CONF_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_configuration_desc failed\n");
		usbsm_cleanup(softstate);
		return DDI_FAILURE;
	}
	usbsm_debugprint("USB config num is %d\n",
			 softstate->usb_config.bConfigurationValue);
	softstate->usb_interfacenum=usb_get_interface_number(dip);
	usbsm_debugprint("USB interface num is %d\n",
				softstate->usb_interfacenum);

	/* If interfacenum==-1, set it to 0: we are responsible for full dev*/
	if(softstate->usb_interfacenum==-1){
		usbsm_debugprint("  (resetting interface num to 0)\n");
		softstate->usb_interfacenum=0;
	}

	/* Fill out "interface" descriptor */
	if(usb_parse_interface_descr(usb_rawconf,usb_raw_len,
				    softstate->usb_interfacenum,
				    0,
				    &softstate->usb_interface_d,
				    USB_IF_DESCR_SIZE) != USB_IF_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_interface_desc failed");
		usbsm_cleanup(softstate);
		return DDI_FAILURE;
	}

	usbsm_debugprint("number of endpoints = %d\n",
			 softstate->usb_interface_d.bNumEndpoints);

	/* Fill out "endpoint" descriptor */
	if(usb_parse_endpoint_descr(usb_rawconf, usb_raw_len,
					softstate->usb_interfacenum,
					0,/* alt interfacenum*/
					0,/*endpoint index num*/
					&softstate->endpoint,
					USB_EPT_DESCR_SIZE) !=
						USB_EPT_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_endpoint_descr failed\n");
		usbsm_cleanup(softstate);
		return DDI_FAILURE;
	}

	/* FYI: type3==USB_EPT_ATTR_INTR */
	usbsm_debugprint("endpoint %d: type %d",
			 0,
			 softstate->endpoint.bmAttributes);

	usbsm_debugprint("INTERFACE Class=%x, subclass=%x\n",
			 softstate->usb_interface_d.bInterfaceClass,
 			 softstate->usb_interface_d.bInterfaceSubClass);

	if(usb_pipe_bulk_transfer_size(dip,&softstate->max_transfer_size)!=USB_SUCCESS)
	{
		cmn_err(CE_NOTE,"usbsm: bulk_transfer_size detect failed");
	}
	usbsm_debugprint(" bulk_maxtransfer sizeval=%u",softstate->max_transfer_size);
	usbsm_debugprint(" maxtransfer sizeval=%u",
	                  softstate->endpoint.wMaxPacketSize);


#ifdef STRINGSAMPLE
	/* We dont need to do this, because ddi_dev_report() does
	 * the same thing. See /var/adm/messages, Not console.
	 * But this is here in case you want an
	 * example of pulling string info from the USB dev
	 */
	{
		usb_device_descr_t * usb_dev_d;
		char devname[20];
		usb_dev_d=usb_get_dev_descr(dip);
		if(usb_dev_d!=NULL){
			usb_get_string_descriptor(dip,USB_LANG_ID,
						  usb_dev_d->iProduct,
						  devname,19);
			devname[19]='\0';
			usbsm_debugprint("ProductName is %s\n",devname);
			
		}
		
	}
#endif /* STRINGSAMPLE */

	/* This will create a device node in the appropriate place
	 * under /devices/pci[*]/XXXXXX
	 *
	 * To get an automatic matching link in /dev/XXXX, you need to 
	 * add your own line to /etc/devlink.tab
	 */

	/* XXX I should really check the dev type, and change the
	 * name based on that stuff
	 */
	ddi_create_minor_node(dip, DEVNODE_NAME, S_IFCHR,
			  INST_TO_NODENUM(instance),
			  DEVNODE_TYPE, 0);

	/* There are no PCI-style interrupts, since we dont get to directly
	 * connect to a device.
	 * but an interrupt handler of sorts gets set up in open()
	 */


	ddi_report_dev(dip);

	/* Dont forget to make ddi_remove_minor_node match in detach() */

	return DDI_SUCCESS;

	/* if you add anything to this routine, make ultra-paranoid
	 * checks on return values and return (DDI_FAILURE)
	 * if ANYTHING doesnt go completely 100% okay
	 */
}


/* called by usbsm_wput */
static void usbsm_ioctl(queue_t *q,mblk_t *mblk){
	struct iocblk * ioctlptr;
	struct usbsm_softstate *softstate;
	int buttons = 5;

	ioctlptr=(struct iocblk *)mblk->b_rptr;
	softstate=(struct usbsm_softstate *)q->q_ptr;

	if(softstate==NULL){
		printf("ERR: softstate not set, in usbsm_ioctl\n");
		ioctlptr->ioc_rval=0;
		ioctlptr->ioc_error=0;
		mblk->b_datap->db_type=M_IOCACK;
		qreply(q,mblk);
		return;
	}

	printf("DEBUG: ioctl(%d) called\n", ioctlptr->ioc_cmd);

	switch(ioctlptr->ioc_cmd){
		case VUIDSFORMAT:
			printf("DEBUG: doing VUIDSFORMAT\n");
			softstate->vuid_format=1;
			ioctlptr->ioc_rval=0;
			ioctlptr->ioc_error=0;
			mblk->b_datap->db_type=M_IOCACK;
			qreply(q,mblk);

			return;

		case VUIDGFORMAT:
			printf("DEBUG: doing VUIDGFORMAT now\n");
			mblk->b_cont=allocb(4, BPRI_HI);
			bcopy(&softstate->vuid_format,
			      mblk->b_cont->b_wptr, 4);
			mblk->b_cont->b_wptr +=4;

			ioctlptr->ioc_count=4;
			ioctlptr->ioc_rval=0;
			ioctlptr->ioc_error=0;
			mblk->b_datap->db_type=M_IOCACK;
			qreply(q,mblk);
			return;

#ifdef MSIOBUTTONS
		case MSIOBUTTONS:
			usbsm_debugprint("Doing MSIOBUTTONS ioctl now\n");
			mblk->b_cont=allocb(sizeof(buttons), BPRI_HI);
			bcopy(&buttons,
			      mblk->b_cont->b_wptr, sizeof(buttons));
			mblk->b_cont->b_wptr += sizeof(buttons);

			ioctlptr->ioc_count=sizeof(buttons);
			ioctlptr->ioc_rval=0;
			ioctlptr->ioc_error=0;
			mblk->b_datap->db_type=M_IOCACK;
			qreply(q,mblk);
			return;
#endif

		default:
			printf("unrecognized ioctl %d\n",ioctlptr->ioc_cmd);
			ioctlptr->ioc_rval=0;
			ioctlptr->ioc_error=EINVAL;
			mblk->b_datap->db_type=M_IOCNAK;
			qreply(q,mblk);
			return;
	}
}

/* This handles (w)rites to the device.
 * While we dont actually send anything TO the device.. we have
 * to support the user sending ioctls TO US.
 * This is the way STREAMS drivers handle that sort of thing.
 *
 */
static	int	usbsm_wput(queue_t *q,mblk_t *mblk){
	switch(mblk->b_datap->db_type){
		case M_IOCTL:
			usbsm_ioctl(q, mblk);
			return 0;

		/* we cant pass this one on: it crashes us */
		case M_FLUSH:

			/* for writequeue*/
			if(*mblk->b_rptr & FLUSHW){
				/*flushq(q,FLUSHDATA);*/
				*mblk->b_rptr &= ~FLUSHW;
			}

			if(*mblk->b_rptr & FLUSHR){
				flushq(RD(q),FLUSHDATA);
				qreply(q,mblk);
			} else {
			
				freemsg(mblk);
			}
			return 0;

		default:
			/* this probably gets dropped intoa black hole,
			 * by the system. but this is the "proper"
			 * way to default things
			 */
			(void)putnext(q,mblk);
	}

	return 0;
}


/* usbsm_rput:   precursor to the usbsm_rsrv routine.
 * We put stuff on OUR 'STREAM' queue here.
 * Then the rsrv routine moves the stuff over to the UPSTREAM queue,
 * aka the 'stream head', which the user talks to.
 */
static int
usbsm_rput(queue_t *q,mblk_t *mp)
{
	if(debug_level >=2){
		usbsm_debugprint("^rput called. Placing %p on our queue\n",mp);
	}

	putq(q,mp);
	return 0;
}


/* STREAMS 'service' proceedure, for the read queue.
 * called by the kernel STREAMS scheduler.
 * Basically, just takes all the mouse events we have queued up internally,
 * and shoves them up the stream towards user-level.
 */
static int
usbsm_rsrv(queue_t *q)
{
	mblk_t *mq;

	if(debug_level>=2){
		usbsm_debugprint("^rsrv called\n");
	}

	while((mq=getq(q)) != NULL){
		if(!canputnext(q)){
			putbq(q,mq);
			break;
		}
		putnext(q,mq);
	}
	return 0;
}


static int
usbsm_open(queue_t *rq,dev_t *dev, int openflags, int sflag, cred_t *credp){
	int instance=DEVT_TO_INST(*dev);
	usb_pipe_policy_t policy,*policyptr;
	struct usbsm_softstate *softstate;

	ASSERT(rq); /* bit of sun paranoia */

	usbsm_debugprint("usbsm: open for instance %d\n", instance);

	/* We dont handle any special stream options
	 * like MODOPEN or CLONEOPEN
	 */
	if (sflag){
		cmn_err(CE_WARN,"Cannot open with STREAM flags set\n");
		return EINVAL;
	}
/* MUST allow openforwrite, 'cause Xserver does that. bizzare */
/*****
	if (openflags & FWRITE){
		usbsm_debugprint("Cannot open device for write\n");
		return EINVAL;
	}
*****/

	softstate=ddi_get_soft_state(usbsm_glob_soft_handle, instance);
	if(softstate==NULL){
		/* This is NOT technically a failure: it is sometimes expected.
		 * It is called "deferred attach"
		 */
		return (ENXIO);
	}

	policyptr=&policy;
	bzero(policyptr,sizeof(policy));

	policyptr->pp_version=USB_PIPE_POLICY_V_0;
	policyptr->pp_timeout_value=USB_PIPE_TIMEOUT;
	policyptr->pp_callback_arg=(void*)softstate;/* same for ALL callbacks*/

	policyptr->pp_callback=usbsm_pipe_callback;
	policyptr->pp_exception_callback=usbsm_exception_callback;

	/* we do not set these, because there are no ISOC endpoints */
	/*
	policyptr->pp_isoc_callback=usbsm_isoc_callback;
	policyptr->pp_isoc_exception_callback=usbsm_isoc_exception_callback;
	*/

	/* headerfile seems to hint we shold be able to use
	 * max_transfer_size==0. but that dont work.
	 */
	policyptr->pp_periodic_max_transfer_size=
	            softstate->endpoint.wMaxPacketSize;

	mutex_enter(&softstate->instance_mutex);
	if(softstate->openflags!=0){
		usbsm_debugprint("usbsm: device already opened\n");
		mutex_exit(&softstate->instance_mutex);
		return (EBUSY);
	}
	softstate->openflags=USBMS_OPEN_HELD;
	mutex_exit(&softstate->instance_mutex);

	if(usb_pipe_open(softstate->dip,&softstate->endpoint, policyptr,
			 USB_FLAGS_SLEEP|USB_FLAGS_SHORT_XFER_OK,
			 &softstate->usb_pipe_handle) !=USB_SUCCESS){
		cmn_err(CE_WARN,"usb_pipe_open failed");
		softstate->openflags=0;
		return (ENXIO);
	}

	/* Not sure whether this should be after, or before, start_polling */
	softstate->rq=rq;

	/* Waitaminit.. why do I need to set a "policy", if I can
	 * specify a callback thing here? !!
	 */
	if(usb_pipe_start_polling(softstate->usb_pipe_handle,
				  USB_FLAGS_SLEEP|USB_FLAGS_SHORT_XFER_OK,
				  usbsm_test_callback,NULL) != USB_SUCCESS)
	{
		cmn_err(CE_WARN,"start_polling FAILED");
		softstate->openflags=0;
		return ENXIO;
	}

	usbsm_debugprint("usbsm_open called start_polling successfully\n");

	/* Kinda redundant since we can just do ddi_get_soft_state,
	 * but this is the "traditional" streams way of doing things.
	 * Stream entrypoints can then access q->q_ptr to get softstate
	 */
	rq->q_ptr = WR(rq)->q_ptr = softstate;
	qprocson(rq);

	

   	return 0;
}


static int usbsm_close(queue_t *q, int flag, cred_t *cred_p){
	struct usbsm_softstate *softstate;

	/* Okay, extreme paranoia for my first STREAMS dev */
	if(q==NULL) return ENXIO;
	if(q->q_ptr==NULL) return ENXIO;

	qprocsoff(q);

	/* XXX in theory, might want to do additional stuff like
	 *  quntimeout(), qunbufcall(), .... ?
	 */
	softstate=(struct usbsm_softstate*)q->q_ptr;

   	usbsm_debugprint("usbsm: close called\n");
	mutex_enter(&softstate->instance_mutex);
	if(softstate->openflags!= USBMS_OPEN_HELD){
		cmn_err(CE_WARN,"ERROR: usbsm_close: pipe not open");
		mutex_exit(&softstate->instance_mutex);
		return ENXIO;
		
	}
	softstate->openflags= USBMS_OPEN_ENDING;
	mutex_exit(&softstate->instance_mutex);

	usb_pipe_stop_polling(softstate->usb_pipe_handle,
			      USB_FLAGS_SLEEP,NULL,NULL);

	if(usb_pipe_close(&softstate->usb_pipe_handle,
			  USB_FLAGS_SLEEP,NULL,NULL ) != USB_SUCCESS)
	{
		usbsm_debugprint("ERROR: usb_pipe_close failed?");
		return ENXIO;
	}
	softstate->usb_pipe_handle=NULL;
	softstate->openflags=0;

	q->q_ptr=WR(q)->q_ptr=NULL;

   	return (0);
}

static int usbsm_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
   	int instance;
	struct usbsm_softstate *softstate;
   	instance=ddi_get_instance(dip);

	usbsm_debugprint("usbsm: detach for instance %d\n", instance);

	softstate=ddi_get_soft_state(usbsm_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"usbsm_detach failed: cannot get softstate?");
		return DDI_FAILURE;
	}

	usbsm_cleanup(softstate);

	ddi_remove_minor_node(dip, DEVNODE_NAME);

	return (DDI_SUCCESS);
		     
}
